import PotentialInfiteLoopError from "./errors/potentialInfinteLoopError";
import TwoManyInstructionError from "./errors/twoManyInstructionError";
import UnknownInstructionOrCellId from "./errors/UnknownInstructionOrCellId";
import InputReader from "./inputReader";
import ExecutionResult from "./instructions/executionResult";
import { executeInstruction, getInstructionLength } from "./instructions/instructionsExecutor";
import { Cells, Register } from "./registers";


export default class Sprint {
  maxInstructionCount: number;
  cells: Cells;
  inputReader: InputReader;
  pc: number;
  insturctionsExecuted: number;
  isHalted: boolean = false;
  stepExecutionResult: ExecutionResult[];

  constructor(
    maxInstructionCount: number,
    registers: Cells,
    inputReader: InputReader,
    programCounter: number = 1,
    instructionExecuted: number = 0,
    stepExecutionResult: ExecutionResult[] = [],
    isHalted: boolean = false
  ) {
    this.maxInstructionCount = maxInstructionCount;
    this.cells = registers;
    this.pc = programCounter;
    this.insturctionsExecuted = instructionExecuted;
    this.stepExecutionResult = stepExecutionResult;
    this.isHalted = isHalted;
    this.inputReader = inputReader;
  }

  static getInstance(
    maxInstructionCount: number,
    cellCount: number,
    code: string,
    inputReader: InputReader
  ) {
    return new Sprint(
      maxInstructionCount,
      Sprint.parse(
        updateLablesWithCellPositions(code), cellCount),
      inputReader
    );
  }

  nextInstructionLength() {
    if (this.isHalted) {
      return 0;
    }
    return getInstructionLength(this.cells, this.pc);
  }

  executeNext() {
    if (this.insturctionsExecuted > this.maxInstructionCount) {
      throw new PotentialInfiteLoopError(this.maxInstructionCount);
    }
    if (this.isHalted) {
      return;
    }
    const executionResult = executeInstruction(this.cells.copy(), this.pc, this.inputReader);

    return new Sprint(
      this.maxInstructionCount,
      executionResult.cells.copy(),
      this.inputReader,
      executionResult.pc,
      this.insturctionsExecuted + 1,
      [...this.stepExecutionResult, { ...executionResult, previousCells: this.cells, previousPc: this.pc }],
      executionResult.isHalted
    );
  }
  execute() {
    while (!this.isHalted) {
      if (this.insturctionsExecuted > this.maxInstructionCount) {
        throw new PotentialInfiteLoopError(this.maxInstructionCount);
      }
      const executionResult = executeInstruction(this.cells.copy(), this.pc, this.inputReader);

      this.cells = executionResult.cells.copy();
      this.pc = executionResult.pc;
      this.insturctionsExecuted++;
      this.stepExecutionResult.push(
        {...executionResult, previousCells: this.cells, previousPc: this.pc}
      );
      this.isHalted = executionResult.isHalted;
    }
    return this;
  }
  executePrevious() {
    if (this.pc === 1) {
      return;
    }
    const lastExecutionResult = this.stepExecutionResult[this.stepExecutionResult.length - 1];

    return new Sprint(
      this.maxInstructionCount,
      lastExecutionResult.previousCells.copy(),
      this.inputReader,
      lastExecutionResult.previousPc,
      this.insturctionsExecuted - 1,
      this.stepExecutionResult.slice(0, this.stepExecutionResult.length - 1)
    );
  }

  static parse(code: string, cellsCount: number) {
    const cells: Register[] = new Array(cellsCount).fill(0)
      .map((_, i) => ({ id: i, value: undefined }));

    const tokens = code.split("\n")
      .flatMap((line) => line.trim().split(" "))
      .map(Sprint.toInt);

    if (tokens.length > cellsCount) {
      throw new TwoManyInstructionError(tokens.length, cellsCount);
    }

    tokens.forEach((value, i) => {
      const cell = cells[i];
      if (cell) cell.value = value;
    });

    return new Cells(cellsCount, cells);
  }

  static toInt(token: string) {
    const int = parseInt(token);
    if (isNaN(int)) {
      throw new UnknownInstructionOrCellId(token);
    }
    return int;
  }
}

const ignoreComment = (line: string) => {
  if (line.includes(";")) {
    return line.substring(0, line.indexOf(";"));
  }
  return line;
};

const addLableIfNotPresent = (
  labels: Map<string, number>,
  label: string,
  position: number,
  failOnDuplicate: boolean,
) => {
  if (labels.get(label) && failOnDuplicate) {
    throw new Error(`Lable ${label} already declared for cell number ${labels.get(label)}`);
  }
  labels.set(label, position);
};

type Context = {
    labels: Map<string, number>,
    cellsCount: number,
    tokensWithOutLableDeclaration: string[],
}

const getLabelsAndTokens = (rawCode: string, failOnDuplicate: boolean = true) => {
  const context: Context = {
    labels: new Map(),
    cellsCount: 1,
    tokensWithOutLableDeclaration: []
  };
  rawCode.split("\n")
    .map(ignoreComment)
    .filter(line => line !== "")
    .reduce((acc, line) => {
      const tokens = line.replaceAll(":", ": ").split(/\s+/).filter(token => token !== "");
      tokens.forEach(token => {
        if (token.includes(":")) {
          const label = token.replace(":", "").trim();
          addLableIfNotPresent(acc.labels, label, acc.cellsCount, failOnDuplicate);
        }
        else {
          acc.cellsCount += 1;
        }
      });
      tokens.filter(token => !token.includes(":")).forEach(token => acc.tokensWithOutLableDeclaration.push(token));
      return acc;
    }, context);
  return context;
};

const updateLabelsWithCellPositions = (context: Context) => {
  for (const entry of context.labels.entries()) {
    const updatedTokens = context.tokensWithOutLableDeclaration
      .map(token => token === entry[0] ? (entry[1] || "").toString() : token);
    context.tokensWithOutLableDeclaration = updatedTokens;
  }
  return context;
};

const validateNoUndeclaredVariableIsUsed = (context: Context) => {
  context.tokensWithOutLableDeclaration.forEach((token, index) => {
    if (isNaN(parseInt(token))) {
      throw new Error(`Use of undeclared label ${token} at cell number ${index + 1}`);
    }
  });
  return context;
};

const updateLablesWithCellPositions = (rawCode: string) => {
  const context = getLabelsAndTokens(rawCode);
  updateLabelsWithCellPositions(context);
  validateNoUndeclaredVariableIsUsed(context);
  return context.tokensWithOutLableDeclaration.join(" ");
};
